function h=magicslidebox(varargin)
% MAGICSLIDEBOX(pos,varnames,ranges,fieldname,listener,defaults)
% Adds a magic slide box to a gui, in the specified position (pos) with the variable names (varnames).
% Includes slide bars for each variable over specified ranges (ranges) with optional default values
% (defaults). If defaults is omitted, the mean value of the range is used.
%
% Position is specified in units of current figure.
%
% Optional input arguement fieldname is the field of handles (guidata) where all handles of objects
% created here and other relevant info will be stored. Make this different for each magic slide box
% you create.
%
% Sliding the slide bar after creation will also execute the function in listener. example:
% @magiclisten
%
% RAVE Rapid Analysis and Visualization Environment
% Copyright (C) 2010 Georgia Tech Research Corporation
% ALL RIGHTS RESERVED
% 
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; version 2
% of the License.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% Written by: Matthew Daskilewicz mdaskilewicz@asdl.gatech.edu



if nargin<4
    error('Not enough input arguements')
elseif nargin==4
    defaults=[];
    pos=varargin{1};
    varnames=varargin{2};
    ranges=varargin{3};
    fieldname=varargin{4};
    listener='';
elseif nargin==5
    defaults=[];
    pos=varargin{1};
    varnames=varargin{2};
    ranges=varargin{3};
    fieldname=varargin{4};
    listener=varargin{5};
else
    defaults=varargin{6};
    pos=varargin{1};
    varnames=varargin{2};
    ranges=varargin{3};
    fieldname=varargin{4};
    listener=varargin{5};
end

%get guidata from current gui:
handles=guidata(gcf);

%check that none of the fields being used here already exist in the gui:
if isfield(handles,fieldname)
%    warning('Some of the field names needed to create magic box already exist in handles. They will be overwritten.')
end

%format position:
axpos=[pos(1) pos(2) pos(3)-18 pos(4)];
slpos=[pos(1)+pos(3)-17 pos(2)-1 17 pos(4)+1];

%create the axes:
handles.(fieldname).ax=axes('units',get(gcf,'units'),'position',axpos,...
    'xlimmode','manual','ylimmode','manual','xlim',[0 1],'ylim',[0 1],...
    'xtickmode','manual','ytickmode','manual','xtick',[],'ytick',[],'box','on',...
    'xcolor',[.7 .7 .7 ],'ycolor',[.7 .7 .7]);

%create the slidebar:
handles.(fieldname).slider=uicontrol('units',get(gcf,'units'),'position',slpos,...
    'style','slider');

nvars=numel(varnames);
%handles.ranges=repmat([1 9],nvars,1);

ndisp=floor(pos(4)/14.7); %number of rows of text that will fit in the box

%make box:
if nvars>ndisp
    for t=1:nvars
        text(0.38,t/ndisp,varnames{end-t+1},'horizontalalignment','right','clipping','on','verticalalignment','top');

        set(handles.(fieldname).slider,'max',nvars/ndisp-1,'value',nvars/ndisp-1);
        set(handles.(fieldname).ax,'ylim',[nvars/ndisp-1 nvars/ndisp]);
        range=get(handles.(fieldname).slider,'max')-get(handles.(fieldname).slider,'min');
        set(handles.(fieldname).slider,'sliderstep',[1/(ndisp*range) (1-1/ndisp)/range])

        L(t)=line([0.43 .73],[t/ndisp-.03 t/ndisp-.03],'linewidth',1,'color',[0 0 0],'marker','+');

        l(t)=line(0.58,t/ndisp-.03,'linestyle','none','markerfacecolor',[.9 .9 .9],'marker','o',...
            'markeredgecolor',[0 0 0],'buttondownfcn',{@slidebarclick,fieldname},'userdata',nvars-t+1);

        handles.(fieldname).valtexts(nvars-t+1)=text(0.78,t/ndisp,num2str(mean(ranges(t,:))),'horizontalalignment','left','clipping','on','verticalalignment','top');
    end

else %small amount of variables... don't need slider
    for t=1:nvars
        text(0.38,(t+ndisp-nvars)/ndisp,varnames{end-t+1},'horizontalalignment','right','clipping','on','verticalalignment','top');

        set(handles.(fieldname).slider,'enable','off','visible','off')
        set(handles.(fieldname).ax,'position',pos)

        L(t)=line([0.43 .73],[(t+ndisp-nvars)/ndisp-.025 (t+ndisp-nvars)/ndisp-.025],...
            'linewidth',1,'color',[0 0 0],'marker','+');

        l(t)=line(0.58,(t+ndisp-nvars)/ndisp-.025,'linestyle','none','markerfacecolor',[.9 .9 .9],'marker','o',...
            'markeredgecolor',[0 0 0],'buttondownfcn',{@slidebarclick,fieldname},'userdata',nvars-t+1);

        handles.(fieldname).valtexts(nvars-t+1)=text(0.78,(t+ndisp-nvars)/ndisp,num2str(mean(ranges(t,:))),'horizontalalignment','left','clipping','on','verticalalignment','top');
    end

end

h=[handles.(fieldname).ax handles.(fieldname).slider handles.(fieldname).valtexts L l];

%store ranges for future use.
handles.(fieldname).ranges=ranges;

%on/off switches:
handles.(fieldname).onoff=false(nvars,1);

guidata(gcf,handles);


%% CALLBACKS:
function magicslider_Callback(hObject, eventdata, handles)
ytop=get(handles.(fieldname).slider,'value');
set(handles.(fieldname).ax,'ylim',[ytop ytop+1]);


function slidebarclick(hObject,eventdata,fieldname)

set(gcf,'windowbuttonmotionfcn',{@slidebarslide,fieldname})
set(gcf,'windowbuttonupfcn',@slidebardone)

%double click to turn variables on/off:
if strcmp(get(gcf,'selectiontype'),'open') %double click
    handles=guidata(gcf);
    if handles.(fieldname).onoff(get(gco,'userdata')) %if on, turn off
        set(gco,'markerfacecolor',[.9 .9 .9])
        handles.(fieldname).onoff(get(gco,'userdata'))=0;
        guidata(gcf,handles);
    else
        cmap=getexcelcolors;
        thisvar=get(gco,'userdata');
        set(gco,'markerfacecolor',cmap(thisvar,:))
        handles.(fieldname).onoff(get(gco,'userdata'))=1;
        guidata(gcf,handles);
    end
end


function slidebarslide(hObject,eventdata,fieldname)
handles=guidata(gcf);
clicked=get(handles.(fieldname).ax,'currentpoint');
val=clicked(1,1);

%move marker
if val>0.73
    val=0.73;
elseif val<0.43
    val=0.43;
end
set(gco,'xdata',val)

%update text values
nvars=size(handles.(fieldname).ranges,1);
thisvar=get(gco,'userdata');
range=handles.(fieldname).ranges(nvars-thisvar+1,:);
newval=range(1)+((val-0.43)/0.3)*(range(2)-range(1));

set(handles.(fieldname).valtexts(thisvar),'string',num2str(newval));

function slidebardone(varargin)
set(gcf,'windowbuttonmotionfcn','')
set(gcf,'windowbuttonupfcn','')



